# Copyright (C) 2003 by Dr. Dieter Maurer <dieter@handshake.de>
# D-66386 St. Ingbert, Eichendorffstr. 23, Germany
#
#			All Rights Reserved
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose and without fee is hereby granted,
# provided that the above copyright notice and this permission
# notice appear in all copies, modified copies and in
# supporting documentation.
# 
# Dieter Maurer DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL Dieter Maurer
# BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL
# DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR
# PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER
# TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
"""analyse object for methods (names, arguments, documentation) and
permissions."""

# necessary to activate the inspect patch -- do this early!
import InspectMonkeyPatch

from sys import modules
from inspect import getsourcefile, getsourcelines
from os.path import join, dirname

from Globals import InitializeClass
from AccessControl.SecurityInfo import ClassSecurityInfo, \
     ACCESS_PUBLIC, ACCESS_PRIVATE, ACCESS_NONE
from OFS.Folder import Folder
from DocumentTemplate.DT_String import String as DT_String
from DocumentTemplate.DT_Util import safe_callable

from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
from Products.PythonScripts.PythonScript import PythonScript

from analyse import Doc


#######################################################################
# Permissions
DocFinderAddInstance= 'DocFinder: add instance'
DocFinderDocumentObject= 'DocFinder: document object'
DocFinderShowCode= 'DocFinder: show code'

_mdict= globals()

#######################################################################
## Class
class DocFinder(Folder):
  '''Persistent 'DocFinder' instance.

  'DocFinder' can be customized via its 'Templates' and 'Properties' tabs.

  Its '__call__' method shows documentation via 'showTree'.
  The '__call__' signature is modelled in such a way that
  the 'DocFinder' instance can be used as a direct replacement
  of the 'showDocumentation' DTML method of older 'DocFinder'
  versions.
  '''

  meta_type= "DocFinder"

  security= ClassSecurityInfo()

  security.declareProtected(DocFinderDocumentObject, 
                            '__call__', # not sure whether this is effective
                            'analyseDocumentation',
                            'default_showTree',
                            )

  security.declareProtected(DocFinderShowCode, 
                            'getCodeFor',
                            'showCodeFor',
                            'default_showCode',
                            )

  security.declarePublic('listDocumentationTypes',
                         )

  manage_options= (
    (
    { 'label':'Templates', 'action':'manage_main', },
    )
    + Folder.manage_options[1:]
    )

  _properties= (
    Folder._properties + (
    {'id':'DocType__', 'type':'selection', 'select_variable':'listDocumentationTypes', 'mode':'w'},
    )
    )
  DocType__= 'scripter'

  def listDocumentationTypes(self):
    return ('programmer','scripter','web')

  def analyseDocumentation(self,object,type,method_re):
    '''return the 'Doc' decribing the documentation for *object* of type *type*,
    potentially restricted to methods given by regular expression *methods_re*.
    '''
    return Doc(object,type,method_re).__of__(self)

  def getCodeFor(self,classOrMethod):
    '''return the code for *classOrMethod* (a name)
    as '(sourcefile,startline,endline,modulesource)'.'''
    # determine object -- we do not use import for security reasons
    # we want to show code only, if it is incorporated in Zope
    ncs= classOrMethod.split('.'); path= ''; obj= None
    for nc in ncs:
      path= (path and path + '.') + nc
      if hasattr(obj,nc): obj= getattr(obj,nc)
      else: obj= modules.get(path)
      if obj is None: break
    if obj is None:
      raise ValueError('not found: %s' % classOrMethod)
    try:
      sf= getsourcefile(obj)
      lines,startline= getsourcelines(obj)
      endline= startline+len(lines)
    except TypeError:
      raise ValueError('not a class or method: %s' % classOrMethod)
    except IOError:
      sf= startline= endline= None
    if sf is not None: source= open(sf).read()
    return sf,startline,endline,source

  def showCodeFor(self,classOrMethod):
    '''show code for *classOrMethod* (a name) through delegation to 'showCode'.'''
    template= self._resolve('showCode')
    return template(codeFor= classOrMethod,
                    codeInfo= self.getCodeFor(classOrMethod))

  ## emulate the old DMTL Method 'showDocumentation'
  isDocTemp= 1
  func_code= DT_String.func_code
  func_defaults= DT_String.func_defaults

  def __call__(self,client=None,REQUEST=None,**kw):
    '''the 'showDocumentation' emulation.'''
    if REQUEST is None:
      REQUEST= getattr(self,'REQUEST',None)
    if REQUEST is None: REQUEST= {}
    if client is None:
      client= kw.get('ObjToDoc__') or REQUEST.get('ObjToDoc__')
    if client is None:
      this= REQUEST.get('this')
      if safe_callable(this): this= this()
      if this is None:
        raise ValueError('no object specified')
      client= this
    doctype= kw.get('DocType__') or REQUEST.get('DocType__') or self.DocType__
    method_re= kw.get('DocMethodRe__') or REQUEST.get('DocMethodRe__')
    template= self._resolve('showTree')
    return template(DocInfo=self.analyseDocumentation(client,doctype,method_re),
                    **kw)

  ## default templates
  default_showCode= PageTemplateFile('zpt/showCode',_mdict)
  default_showTree= PageTemplateFile('zpt/showTree',_mdict)


  ## auxiliaries
  def _resolve(self,name):
    '''return name attribute (maybe acquired) or 'default_'*name*.'''
    obj= getattr(self,name,None)
    if obj is None: obj= getattr(self,'default_' + name)
    return obj

  # prevent acquisition
  index_html= None

InitializeClass(DocFinder)


###################################################################
## 'document' script
_document='''
r= container.REQUEST
U= r.URL
URL= '%s/%s' % (U,'/'.join(traverse_subpath))
r.set('URL',URL)
BURL= '%s/%s' % (U,'/'.join(traverse_subpath[:-1]))
r.RESPONSE.setBase(BURL)
return container.showDocumentation(
   container.restrictedTraverse(['']+traverse_subpath),
   r,
   )
'''


###################################################################
## Constructors
addDocFinderForm= PageTemplateFile('zpt/addForm',_mdict)
def addDocFinder(self,id,RESPONSE=None):
  '''add 'DocFinder' with *id* in 'self'.'''
  df= DocFinder(); df.id= id
  self._setObject(id,df)
  df= self._getOb(id)
  _addPT(df,'showTree')
  _addPT(df,'showCode')
  document= PythonScript('document')
  document.ZPythonScript_setTitle('document object given by traversal subpath')
  document.ZPythonScript_edit('',_document)
  df._setObject(document.id,document)
  if RESPONSE is not None:
    RESPONSE.redirect('%s/manage_main' % df.absolute_url())
    return ''
  return df

def _addPT(self,name):
  dir= dirname(__file__)
  pt= ZopePageTemplate(name,open(join(dir,'zpt',name+'.zpt')).read())
  self._setObject(name,pt)
